import csv
import shutil
from itertools import groupby
from pathlib import Path

from lokii import Lokii
import unittest

DATA_FOLDER = 'data'


class TestLokii(unittest.TestCase):

    def setUp(self) -> None:
        """
        Before running each test remove data directory if exists
        """
        data_path = Path(DATA_FOLDER)
        if data_path.exists() and data_path.is_dir():
            shutil.rmtree(data_path)

    @classmethod
    def tearDownClass(cls) -> None:
        """
        After all tests are finished remove data directory
        """
        data_path = Path(DATA_FOLDER)
        if data_path.exists() and data_path.is_dir():
            shutil.rmtree(data_path)

    def test_simple_table_should_generate_expected_rows(self):
        """
        Test simple table generation.
        """
        simple_count = 100
        lokii = Lokii(out_folder=DATA_FOLDER, silent=True)

        def col_func(x):
            return x * x

        col_const = 'constant_column'

        simple_table = lokii.table('test_simple') \
            .cols('id', 'col_func', 'col_const') \
            .simple(simple_count, lambda i, r: {
                'id': i,
                'col_func': col_func(i),
                'col_const': col_const,
            })
        lokii.generate()

        # Confirm that lokii generates simple table
        with open(simple_table.outfile) as csv_file:
            simple_rows = list(csv.DictReader(csv_file, delimiter=','))

            # Confirm that correct amount of rows are generated
            self.assertEqual(len(simple_rows), simple_count)
            # Confirm that the generated ids are starting from 1
            self.assertEqual(simple_rows[0]['id'], str(1))
            # Confirm that the generated ids end at count
            self.assertEqual(simple_rows[simple_count - 1]['id'], str(simple_count))

            an_index = int(simple_count / 2)
            # Confirm that the generated row yields result of defined function
            self.assertEqual(simple_rows[an_index]['col_func'], str(col_func(an_index + 1)))
            # Confirm that the generated row yields constant
            self.assertEqual(simple_rows[an_index]['col_const'], str(col_const))

    def test_relation_table_should_generate_expected_rows(self):
        """
        Test relation table generation.
        """
        simple_count = 100
        rel_count = 200
        lokii = Lokii(out_folder=DATA_FOLDER, silent=True)

        simple_table = lokii.table('test_simple') \
            .cols('id', 'col') \
            .simple(simple_count, lambda i, r: {
                'id': i,
                'col': 'col',
            })

        rel_table = lokii.table('test_rel') \
            .cols('id', 'rel_id') \
            .rels(simple_table) \
            .simple(rel_count, lambda i, r: {
                'id': i,
                'rel_id': r['test_simple']['id']
            })

        lokii.generate()

        # Confirm that lokii generates relation table
        with open(rel_table.outfile) as csv_file:
            rel_rows = list(csv.DictReader(csv_file, delimiter=','))

            # Confirm that correct amount of rows are generated
            self.assertEqual(len(rel_rows), rel_count)
            # Confirm that the generated ids are starting from 1
            self.assertEqual(rel_rows[0]['id'], str(1))
            # Confirm that the generated ids end at count
            self.assertEqual(rel_rows[rel_count - 1]['id'], str(rel_count))

            # Confirm that all generated rows contains valid relation id
            for row in rel_rows:
                self.assertTrue(1 <= int(row['rel_id']) <= simple_count)

    def test_product_table_should_generate_expected_rows(self):
        """
        Test product table generation.
        """
        simple_count = 100
        lokii = Lokii(out_folder=DATA_FOLDER, silent=True)

        simple_table = lokii.table('test_simple') \
            .cols('id', 'col') \
            .simple(simple_count, lambda i, r: {
                'id': i,
                'col': 'col',
            })

        multiplier = [1, 2, 3]
        product_table = lokii.table('test_product') \
            .cols('id', 'multiplicand_id', 'multiplier') \
            .multiply(simple_table, lambda i, m, r: {
                'id': i,
                'multiplicand_id': r['test_simple']['id'],
                'multiplier': m
            }, multiplier)

        lokii.generate()

        # Confirm that lokii generates product table
        with open(product_table.outfile) as csv_file:
            product_rows = list(csv.DictReader(csv_file, delimiter=','))

            # Confirm that correct amount of rows are generated
            self.assertEqual(len(product_rows), simple_count * len(multiplier))

            sorted_rows = sorted(product_rows, key=lambda r: r['multiplicand_id'])
            grouped_rows = [list(it) for k, it in groupby(sorted_rows, lambda r: r['multiplicand_id'])]
            for rows in grouped_rows:
                # Confirm that all multiplied rows of simple table has multiplied by all multiplier items
                self.assertEqual(sorted([int(r['multiplier']) for r in rows]), sorted(multiplier))

    def test_table_should_not_write_none_rows(self):
        """
        Test product table generation.
        """
        simple_count = 100
        lokii = Lokii(out_folder=DATA_FOLDER, silent=True)

        simple_table = lokii.table('test_simple') \
            .cols('id', 'col') \
            .simple(simple_count, lambda i, r: {
                'id': i,
                'col': 'col',
            } if i < int(simple_count / 2) + 1 else None)

        multiplier = [1, 2, 3]
        product_table = lokii.table('test_product') \
            .cols('id', 'multiplicand_id', 'multiplier') \
            .multiply(simple_table, lambda i, m, r: {
                'id': i,
                'multiplicand_id': r['test_simple']['id'],
                'multiplier': m
            }, multiplier)

        lokii.generate()

        # Confirm that lokii not writes none rows
        with open(simple_table.outfile) as csv_file:
            simple_rows = list(csv.DictReader(csv_file, delimiter=','))

            # Confirm that correct amount of rows are generated
            self.assertEqual(len(simple_rows), int(simple_count / 2))

        # Confirm that lokii generates product table
        with open(product_table.outfile) as csv_file:
            product_rows = list(csv.DictReader(csv_file, delimiter=','))

            # Confirm that correct amount of rows are generated
            self.assertEqual(len(product_rows), int(simple_count / 2) * len(multiplier))



import math
import random
import time
from csv import DictWriter
from functools import partial
from os import makedirs, path, PathLike
from typing import Union, Dict

from pathos.multiprocessing import ProcessingPool

from .table import Table
from .utils import print_process


class Lokii:

    def __init__(self,
                 out_folder: Union[str, bytes, PathLike] = './data',
                 process_count: int = 8,
                 batch_size: int = 10000,
                 write_buffer_size: int = 50000,
                 index_cache_size: int = 50000,
                 random_cache_size: int = 20000,
                 silent: bool = False,
                 debug: bool = False):
        """
        Generates massive amount of relational mock data.

        :param out_folder: path of output folder for mock data
        """
        self.__out_folder = out_folder
        self._tables: Dict[str, Table] = {}

        self._process_count = process_count
        self._batch_size = batch_size
        self._index_cache_size = index_cache_size
        self._random_cache_size = random_cache_size
        self._silent = silent
        self._debug = debug

    def table(self, name: str) -> Table:
        table = Table(name, path.join(self.__out_folder, name + '.csv'), self._index_cache_size,
                      self._random_cache_size, self._debug)

        self._tables[name] = table
        return self._tables[name]

    def generate(self):
        # Create out folder if not exists
        if not path.exists(self.__out_folder):
            makedirs(self.__out_folder)

        for name, table in self._tables.items():
            print('GEN > {}'.format(name))
            t = time.time()

            table.prepare()
            self.__generate_table(name)

            elapsed_time = time.time() - t
            print('END > {}: {} rows in {:.4f}s\n'.format(name, table.row_count, elapsed_time))

    def __generate_table(self, table_name: str):
        table = self._tables[table_name]

        with ProcessingPool(nodes=self._process_count) as pool:
            with open(table.outfile, 'w+', newline='', encoding='utf-8') as outfile:
                writer = DictWriter(outfile, fieldnames=[name for name in table.columns])
                writer.writeheader()

                # Write defaults
                writer.writerows(table.defaults)
                table.row_count += len(table.defaults)
                table.gen_row_count += len(table.defaults)

                while table.row_count < table.target_count:
                    batch_start = table.row_count
                    batch_end = batch_start + self._batch_size \
                        if batch_start + self._batch_size < table.target_count \
                        else table.target_count

                    if table.is_product:
                        # If it is a product table load index cache of multiplicand
                        table.multiplicand.load_index_cache(
                            math.floor(batch_start / len(table.multiplier)),
                            math.floor(batch_end / len(table.multiplier)))

                    for rel in table.relations:
                        # Load random cache for all relation tables
                        rel.load_random_cache(batch_start / table.target_count)

                    if table.multiplicand:
                        rel_dicts = \
                            [
                                {
                                    table.multiplicand.name: table.multiplicand.get_row(
                                        math.floor(index / len(table.multiplier))),
                                    **{r.name: r.get_rand(index)
                                       for i, r in enumerate(table.relations)}
                                }
                                for index in range(batch_start, batch_end)
                            ]
                    else:
                        rel_dicts = \
                            [
                                {
                                    r.name: r.get_rand(index)
                                    for i, r in enumerate(table.relations)
                                }
                                for index in range(batch_start, batch_end)
                            ]

                    result = pool.map(
                        table.gen_func,
                        [index + 1 for index in range(batch_start, batch_end)], rel_dicts)

                    # Remove none rows
                    result = list(filter(None, result))

                    writer.writerows(result)
                    table.row_count = batch_end
                    table.gen_row_count += len(result)
                    print_process(table.row_count, table.target_count)
        print('')
        # Generation completed purge caches
        if table.is_product:
            # If it is a product table purge multiplicand cache
            table.multiplicand.purge_cache()

        for rel in table.relations:
            # Purge cache for all relation tables
            rel.purge_cache()
